package winterwell.utils.containers;

import java.io.Serializable;
import java.util.AbstractMap;
import java.util.AbstractSet;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

/**
 * A map backed by an array. Fast for small collections, lousy for big ones. Has
 * a convenient constructor. Another nice property is that insertion order is
 * preserved. I.e. keySet() values() and entrySet() all return objects in the
 * order they were added.
 * <p>
 * ArrayMap provides iteration over keys which ALLOWS the map to be edited
 * during the iteration. ie. it saves you writing
 * <code>for(X x : map.keySet().toArray(new X[0]))</code> when editing a map.
 * <p>
 * ArrayMap is not thread safe.
 * 
 * @author Daniel
 * 
 * @param <K>
 * @param <V>
 */
public final class ArrayMap<K, V> extends AbstractMap<K, V> implements
		Serializable, Iterable<K> {
	private static final long serialVersionUID = 1L;

	private final ArrayList<K> keys;
	private final ArrayList<V> values;

	/**
	 * Create a map optimised for small cases.
	 */
	public ArrayMap() {
		this(4);
	}

	/**
	 * Create a map with the given starting capacity.
	 */
	public ArrayMap(int initialSize) {
		keys = new ArrayList<K>(initialSize);
		values = new ArrayList<V>(initialSize);
	}

	public ArrayMap(List<Pair2<K, V>> pList) {
		keys = new ArrayList<K>(pList.size());
		values = new ArrayList<V>(pList.size());
		for (Pair2<K, V> kv : pList) {
			keys.add(kv.first);
			values.add(kv.second);
		}
	}

	/**
	 * Warning: Does not check for duplicates!
	 * 
	 * @param keyValuePairs
	 */
	public ArrayMap(Object... keyValuePairs) {
		assert keyValuePairs.length % 2 == 0;
		keys = new ArrayList<K>(keyValuePairs.length / 2);
		values = new ArrayList<V>(keyValuePairs.length / 2);
		for (int i = 0; i < keyValuePairs.length; i += 2) {
			keys.add((K) keyValuePairs[i]);
			values.add((V) keyValuePairs[i + 1]);
		}
	}

	@Override
	public boolean containsKey(Object key) {
		return keys.contains(key);
	}

	@Override
	public boolean containsValue(Object value) {
		return values.contains(value);
	}

	@Override
	public Set<java.util.Map.Entry<K, V>> entrySet() {
		return new AbstractSet<Entry<K, V>>() {
			@Override
			public Iterator<java.util.Map.Entry<K, V>> iterator() {
				return new Iterator<Entry<K, V>>() {
					MapEntry<K, V> entry = new MapEntry<K, V>();
					int i = 0;

					@Override
					public boolean hasNext() {
						return i < keys.size();
					}

					@Override
					public java.util.Map.Entry<K, V> next() {
						entry.reset(keys.get(i), values.get(i));
						i++;
						return entry;
					}

					@Override
					public void remove() {
						keys.remove(i - 1);
						values.remove(i - 1);
						i--;
					}

				};
			}

			@Override
			public int size() {
				return keys.size();
			}

		};
	}

	@Override
	public V get(Object key) {
		int i = keys.indexOf(key);
		if (i == -1)
			return null;
		return values.get(i);
	}

	public List<Pair2<K, V>> getPList() {
		List<Pair2<K, V>> pList = new ArrayList<Pair2<K, V>>(size());
		for (int i = 0, n = size(); i < n; i++) {
			pList.add(new Pair2<K, V>(keys.get(i), values.get(i)));
		}
		return pList;
	}

	/**
	 * Iterate over the map keys and ALLOW editing within the iteration. This is
	 * a convenience for copy-keys then return an iterator. TODO is this useful
	 * enough, given that Map doesn't provide it?
	 */
	@Override
	public Iterator<K> iterator() {
		return new ArrayList(keys).iterator();
	}

	/**
	 * The keys in this map. This is not modifiable!
	 */
	@Override
	public Set<K> keySet() {
		return Collections.unmodifiableSet(Containers.listAsSet(keys));
	}

	@Override
	public V put(K key, V value) {
		int i = keys.indexOf(key);
		if (i == -1) {
			keys.add(key);
			values.add(value);
			return null;
		}
		return values.set(i, value);
	}

	@Override
	public V remove(Object key) {
		int i = keys.indexOf(key);
		if (i == -1)
			return null;
		keys.remove(i);
		return values.remove(i);
	}

	@Override
	public int size() {
		return keys.size();
	}

	/**
	 * The values in this map. This is not modifiable!
	 */
	@Override
	public List<V> values() {
		return Collections.unmodifiableList(values);
	}
}
