package winterwell.utils.containers;

import java.util.AbstractList;
import java.util.Arrays;
import java.util.Comparator;

/**
 * The top-n items from a collection. This finds the unique top-n - ie there are
 * no duplicates. These are the top-n according to sort order, e.g. A, B, C or
 * 1, 2, 3 (dropping the high numbers). Change the comparator to alter this.
 * 
 * @testedby {@link TopNListTest}
 * @author daniel
 * 
 */
public class TopNList<X> extends AbstractList<X> {

	final Comparator comp;
	/** ordered list of the top n. may contain nulls intially */
	private final Object[] objects;

	private int size = 0;

	public TopNList(int n) {
		this(n, null);
	}

	/**
	 * 
	 * @param n
	 * @param comp
	 */
	public TopNList(int n, Comparator<X> comp) {
		assert n > 0;
		this.comp = comp;
		objects = new Object[n];
	}

	/**
	 * *Maybe* add the element. This breaks the true contract for List, as the
	 * element may not get added. Explicitly call {@link #maybeAdd(Object)} for
	 * preference.
	 */
	@Override
	public boolean add(X element) {
		return maybeAdd(element);
	}

	/**
	 * Code copied from {@link Arrays#binarySearch(Object[], object)} - which
	 * annoyingly doesn't accept a pluggable comparison operator.
	 * <p>
	 * Searches the specified array for the specified object using the binary
	 * search algorithm. The array must be sorted into ascending order.
	 * 
	 * @param a
	 *            the array to be searched
	 * @param key
	 *            the value to be searched for
	 * @return index of the search key, if it is contained in the array;
	 *         otherwise, <tt>(-(<i>insertion point</i>) - 1)</tt>. The
	 *         <i>insertion point</i> is defined as the point at which the key
	 *         would be inserted into the array: the index of the first element
	 *         greater than the key, or <tt>a.length</tt> if all elements in the
	 *         array are less than the specified key. Note that this guarantees
	 *         that the return value will be &gt;= 0 if and only if the key is
	 *         found.
	 * @throws ClassCastException
	 *             if the search key is not comparable to the elements of the
	 *             array.
	 */
	private int binarySearch(X obj) {
		int low = 0;
		int high = size() - 1;
		while (low <= high) {
			int mid = (low + high) >>> 1;
			Object midVal = objects[mid];
			int cmp = comp == null ? ((Comparable) midVal).compareTo(obj)
					: comp.compare(midVal, obj);
			if (cmp < 0) {
				low = mid + 1;
			} else if (cmp > 0) {
				high = mid - 1;
			} else
				return mid; // key found
		}
		return -(low + 1); // key not found.
	}

	@Override
	public X get(int index) {
		return (X) objects[index];
	}

	public boolean maybeAdd(X obj) {
		assert obj != null;
		int i = binarySearch(obj);
		if (i >= 0)
			// already a member of the list
			return false;
		// insert?
		i = -i - 1;
		if (i >= objects.length)
			return false;
		// yes, insert
		Object prev = obj;
		for (int j = i; j < objects.length; j++) {
			Object newPrev = objects[j];
			objects[j] = prev;
			prev = newPrev;
		}
		size++;
		if (size == objects.length + 1) {
			size = objects.length;
		}
		return true;
	}

	@Override
	public int size() {
		return size;
	}

}
