package winterwell.utils.time;

import java.io.Serializable;
import java.util.Calendar;

/**
 * An amount of time, e.g. 5 minutes. This class is immutable. Dts can be
 * positive or negative (positive is more common).
 * 
 * @testedby {@link DtTest}
 * @author Daniel
 * 
 */
public final class Dt implements Comparable<Dt>, Serializable {
	private static final long serialVersionUID = 1L;

	/**
	 * No time at all.
	 */
	public static Dt ZERO() {
		// Tried todo as a constant - but the cyclic ref between Time and Dt
		// caused an intermittent bug with a null unit in ZERO
		assert TUnit.MILLISECOND != null;
		// ZERO
		return new Dt(0, TUnit.MILLISECOND);
	}

	private final double n;
	private final TUnit unit;

	public Dt(double n, TUnit unit) {
		this.n = n;
		this.unit = unit;
		assert unit != null;
	}

	/**
	 * Convenience for a Dt measured in milliseconds.
	 * 
	 * @param millisecs
	 */
	public Dt(long millisecs) {
		this(millisecs, TUnit.MILLISECOND);
	}

	/**
	 * Some rounding may occur if the value is not an integer. The integer part
	 * of a Dt is added using Calendar's field-specific mechanisms. The
	 * fractional part is added as milliseconds.
	 * 
	 * @param cal
	 */
	public void addTo(Calendar cal) {
		// add integer part using Calendar field
		int i = (int) n;
		cal.add(unit.getCalendarField(), i);
		double r = n - i;
		if (r == 0)
			return;
		// add fractional part using millisecs
		int ms = (int) (r * unit.millisecs);
		cal.add(TUnit.MILLISECOND.getCalendarField(), ms);
	}

	/**
	 * Dts are typically positive, eg. +1 hour. This convenience method returns
	 * the opposite negative Dt, eg. -1 hour. (It is currently an error to call
	 * this on a Dt which is already negative)
	 */
	public Dt ago() {
		assert n >= 0 : this;
		return new Dt(-n, unit);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	@Override
	public int compareTo(Dt dt2) {
		long ms = getMillisecs();
		long ms2 = dt2.getMillisecs();
		if (ms == ms2)
			return 0;
		return ms < ms2 ? -1 : 1;
	}

	/**
	 * Convert this Dt to a different unit.
	 * <p>
	 * Uses {@link #divide(Dt)}, which mostly does millisecond arithmetic. But
	 * it has a couple of special cases to handle: weeks-in-the-year
	 * months-in-the-year -- though for integer valued Dts only.
	 * 
	 * @param unit
	 * @return
	 */
	public Dt convertTo(TUnit unit) {
		double n2 = divide(unit.dt);
		return new Dt(n2, unit);
	}

	/**
	 * Return the number of the specified time delta that would fit into this
	 * one e.g. TUnit.MINUTE.getDt().divide(TUnit.SECOND.getDt()) -> 60
	 * <p>
	 * Mostly this does millisecond arithmetic. But it has a couple of special
	 * cases to handle: weeks-in-the-year months-in-the-year -- though for
	 * integer valued Dts only.
	 * 
	 * @param bucketSize
	 */
	public double divide(Dt other) {
		if (n == 0)
			return 0;
		// special case integer arithmetic
		if (unit == TUnit.YEAR && other.n == Math.round(other.n)) {
			if (other.unit == TUnit.MONTH)
				return n * 12 / other.n;
			else if (other.unit == TUnit.WEEK)
				return n * 52 / other.n;
		}
		return (n * unit.millisecs) / (other.n * other.unit.millisecs);
	}

	/**
	 * Equals if the same to the millisecond. So 1 day equals 24 hours
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		if (obj == this)
			return true;
		if (obj.getClass() != Dt.class)
			return false;
		Dt dt = (Dt) obj;
		return getMillisecs() == dt.getMillisecs();
	}

	/**
	 * @return the number of milliseconds represented by this period. This
	 *         rounds if we have a fraction.
	 */
	public long getMillisecs() {
		return Math.round(unit.millisecs * n);
	}

	public TUnit getUnit() {
		return unit;
	}

	/**
	 * E.g. the 5 in "5 minutes"
	 */
	public double getValue() {
		return n;
	}

	@Override
	public int hashCode() {
		return new Double(getMillisecs()).hashCode();
	}

	/**
	 * Compare absolute values. So +1 minute is shorter than -1 hour.
	 * 
	 * @param dt2
	 * @return true if this is the shorter Dt
	 */
	public boolean isShorterThan(Dt dt2) {
		assert dt2 != null;
		return Math.abs(getMillisecs()) < Math.abs(dt2.getMillisecs());
	}

	/**
	 * @param x
	 * @return A new Dt which is x times this Dt
	 */
	public Dt multiply(double x) {
		return new Dt(x * n, unit);
	}

	/**
	 * A new DT which keeps the unit of this dt
	 * 
	 * @param b
	 * @return this + b
	 */
	public Dt plus(Dt b) {
		return new Dt(n + b.divide(unit.dt), unit);
	}

	@Override
	public String toString() {
		return ((float) n) + " " + unit.toString().toLowerCase()
				+ (n != 1 ? "s" : "");
	}

	// public static Dt FOREVER() {
	// // TODO Auto-generated method stub
	// return null;
	// }

}
