package winterwell.utils.time;

import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.TimeZone;

import winterwell.utils.StrUtils;

/**
 * A simple immutable alternative to the built-in mess of Date. This just holds
 * a UTC millisecond count. It cannot be edited, but provides convenient methods
 * for creating new Time objects.
 * <p>
 * Rationale: Date is a booby-trapped mess (years that gain and lose 1900
 * depending on the method?!, mixed zero-indexed and one-indexed properties?!,
 * etc.). {@link GregorianCalendar} isn't too bad, but it is kind of heavyweight
 * for many uses.
 * <p>
 * In places, this class uses Calendar - which is always the
 * {@link GregorianCalendar} with GMT (i.e. no daylight savings are applied!)
 * and 24 hour clock (ie. it uses {@link Calendar#HOUR_OF_DAY}). If you need
 * locale-specific calendar support, you should provide this yourself.
 * <p>
 * Can be sorted - sorts as earliest Time first.
 * 
 * @testedby {@link TimeTest}
 * @author daniel
 * 
 */
public final class Time implements Serializable, Comparable<Time> {
	private static final long serialVersionUID = 1L;

	private static GregorianCalendar getCal(int year, int month, int day,
			int hour, int min, int sec) {
		assert month > 0 && month < 13 : month;
		assert day > 0 && day < 32 : day;
		// get a GMT time zone
		assert TimeUtils._GMT_TIMEZONE != null;
		GregorianCalendar cal = new GregorianCalendar(TimeUtils._GMT_TIMEZONE);
		// freshen it up
		cal.clear();
		cal.set(year, month - 1, day, hour, min, sec);
		// if (year<0) {
		// cal.set(Calendar.ERA, GregorianCalendar.BC);
		// }
		return cal;
	}

	/**
	 * Unix Time code (no. seconds since 1970). This is the standard definition
	 * of time in many computer systems. Here it is expressed in milliseconds
	 * (the Java convention)
	 */
	private final long ut;

	/**
	 * Once upon a time called *right now*.
	 */
	public Time() {
		ut = System.currentTimeMillis();
	}

	public Time(Calendar cal) {
		this.ut = cal.getTime().getTime();
	}

	/**
	 * Create a Time object from a Date object.
	 * 
	 * @param date
	 */
	public Time(Date date) {
		this.ut = date.getTime();
	}

	public Time(double longValue) {
		this((long) longValue);
		// double err = Math.abs(longValue() - longValue);
		// if (err > 100) {
		// Log.report("Rounding error of "+new Dt(err, TUnit.MILLISECOND),
		// Level.WARNING);
		// }
	}

	/**
	 * Note: This does not behave the same as the (confusing) Date constructor.<br>
	 * 
	 * The Time will be midnight GMT on the specified date (i.e. no daylight
	 * savings applied).
	 * 
	 * @param year
	 *            The year, e.g. 2008
	 * @param month
	 *            The month, not zero indexed so January=1. Must be in [1,12]
	 * @param day
	 *            The day-of-month, not zero-indexed. Must be in [1,31]
	 */
	public Time(int year, int month, int day) {
		this(year, month, day, 0, 0, 0);
	}

	/**
	 * Note: This does <i>not</i> behave the same as the (confusing) Date
	 * constructor.
	 * <p>
	 * - zero-indexing is consistently not used (except for hour/min/sec where 0
	 * is a normal value)<br>
	 * - All times are in GMT (no daylight savings are applied) <br>
	 * 
	 * @param year
	 *            The year, e.g. 2008
	 * @param month
	 *            The month, not zero indexed so January=1. Must be in [1,12]
	 * @param day
	 *            , The day-of-month, not zero-indexed. Must be in [1,31]
	 */
	public Time(int year, int month, int day, int hour, int min, int sec) {
		this(getCal(year, month, day, hour, min, sec));
		assert month > 0 && day > 0 : "Not zero-indexed!";
	}

	/**
	 * @param ut
	 *            Unix Time code (no. seconds since 1970). This is the standard
	 *            definition of time in many computer systems. Here it is
	 *            expressed in milliseconds (the Java convention)
	 */
	public Time(long ut) {
		this.ut = ut;
	}

	/**
	 * Uses {@link Date#parse(String)} to parse the string. Use
	 * {@link TimeUtils#parse(String, String)} for other patterns. Behaviour
	 * will be locale-specific. It's best to use something else.
	 * 
	 * @param date
	 * @testedby {@link TimeTest#testTime()}
	 */
	public Time(String date) {
		// Is it a timecode?
		if (date.length() > 10 && date.length() < 24
				&& StrUtils.isJustDigits(date)) {
			this.ut = Long.valueOf(date);
			return;
		}
		long _date = Date.parse(date);
		this.ut = _date;
	}

	/**
	 * Sorts earliest-first
	 */
	@Override
	public int compareTo(Time o) {
		return (ut < o.ut ? -1 : (ut == o.ut ? 0 : 1));
	}

	/**
	 * Return the date formatted in the most common European style. XXX - maybe
	 * not a good idea from the i18n point of view.
	 * 
	 * @return
	 */
	public String ddMMyyyy() {
		return format("dd/MM/yyyy");
	}

	/**
	 * Time difference
	 * 
	 * @param b
	 * @return time from this to b in milliseconds, ie. positive if b is after
	 *         this.
	 * @see #diff(Time, TUnit)
	 */
	public long diff(Time b) {
		long dt = b.ut - ut;
		return dt;
	}

	/**
	 * Time difference: compute the time difference in the specified units. The
	 * dt will be positive if b is after this.
	 * 
	 * @param b
	 * @param unit
	 * @return time from this to b, ie. positive if b is after this
	 */
	public Dt diff(Time b, TUnit unit) {
		long d = diff(b);
		Dt dt = new Dt(d, TUnit.MILLISECOND);
		return unit == TUnit.MILLISECOND ? dt : dt.convertTo(unit);
	}

	/**
	 * Positive if b is after this time.
	 * 
	 * @param b
	 * @return
	 */
	public Dt dt(Time b) {
		long dt = diff(b);
		return new Dt(dt, TUnit.MILLISECOND);
	}

	/**
	 * equals() if same class and identical millisecond timecode.
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Time other = (Time) obj;
		if (ut != other.ut)
			return false;
		return true;
	}

	/**
	 * Less inconvenient way of getting a formatted string. Always format as GMT
	 * time so this works as the inverse operation to TimeUtils.parse()
	 * 
	 * @param format
	 *            e.g. "dd MMM"
	 * @return
	 * @see SimpleDateFormat
	 */
	public String format(String format) {
		SimpleDateFormat df = new SimpleDateFormat(format);
		TimeZone zone = TimeZone.getTimeZone("GMT");
		df.setTimeZone(zone);
		return df.format(getDate());
	}

	/**
	 * @return a new Gregorian calendar set to this time.
	 */
	public GregorianCalendar getCalendar() {
		GregorianCalendar cal = new GregorianCalendar(TimeUtils._GMT_TIMEZONE);
		cal.setTimeInMillis(ut);
		return cal;
	}

	/**
	 * Date is horrible, but useful for inter-operating with other systems.
	 */
	public Date getDate() {
		return new Date(ut);
	}

	/**
	 * not zero indexed
	 */
	public int getDayOfMonth() {
		return getCalendar().get(Calendar.DAY_OF_MONTH);
	}

	/**
	 * Convenience for using {@link GregorianCalendar} to get the hour in 24
	 * hour clock.
	 * 
	 * @return hour, e.g. 23 for 11pm
	 */
	public int getHour() {
		return getCalendar().get(Calendar.HOUR_OF_DAY);
	}

	/**
	 * Convenience for using {@link GregorianCalendar} to get the minutes past
	 * the hour.
	 */
	public int getMinutes() {
		return getCalendar().get(Calendar.MINUTE);
	}

	/**
	 * not zero indexed so January=1
	 */
	public int getMonth() {
		return 1 + getCalendar().get(Calendar.MONTH);
	}

	/**
	 * Convenience for using {@link GregorianCalendar} to get the seconds in the
	 * minuts.
	 */
	public int getSeconds() {
		return getCalendar().get(Calendar.SECOND);
	}

	/**
	 * The UTC value for this Time, ie. milliseconds since Unix time-zero.
	 * <p>
	 * Note: this has the same behaviour as the {@link Date#getTime()} method,
	 * giving drop-in compatibility with Date
	 * 
	 * @return
	 */
	public long getTime() {
		return ut;
	}

	/**
	 * Convenience for using {@link GregorianCalendar} to get the year.
	 * 
	 * @return year, e.g. 2008 Negative for BC (which is different from the
	 *         Calendar class!)
	 */
	public int getYear() {
		GregorianCalendar cal = getCalendar();
		int yr = cal.get(Calendar.YEAR);
		int era = cal.get(Calendar.ERA);
		if (era == GregorianCalendar.BC)
			return -yr;
		return yr;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + (int) (ut ^ (ut >>> 32));
		return result;
	}

	/**
	 * @param t
	 * @return true if t is <em>strictly</em> before this time; false otherwise
	 */
	public boolean isAfter(Time t) {
		return ut > t.ut;
	}

	/**
	 * @param t
	 * @return true if t is <em>strictly</em> after this time; false otherwise
	 */
	public boolean isBefore(Time t) {
		return ut < t.ut;
	}

	/**
	 * Equivalent to {@link #getTime()}
	 */
	public long longValue() {
		return ut;
	}

	/**
	 * Opposite of {@link #plus(Dt)}
	 */
	public Time minus(Dt dt) {
		dt = new Dt(-dt.getValue(), dt.getUnit());
		return plus(dt);
	}

	/**
	 * Convenience for {@link #minus(Dt)}
	 * 
	 * @param n
	 * @param unit
	 * @return
	 */
	public Time minus(int n, TUnit unit) {
		return minus(new Dt(n, unit));
	}

	/**
	 * Convenience for {@link #minus(Dt)}
	 */
	public Time minus(TUnit dt) {
		return minus(dt.dt);
	}

	/**
	 * @param dt
	 * @return A new object that is time+dt. Uses Calendar to step neatly.
	 */
	public Time plus(Dt dt) {
		GregorianCalendar cal = getCalendar();
		dt.addTo(cal);
		return new Time(cal);
	}

	/**
	 * Add on more time. Uses Calendar to perform integer addition. E.g. adding
	 * 12 months = 1 year
	 * 
	 * @param n
	 *            Can be negative to subtract some time
	 * @param unit
	 * @return a new Time object
	 */
	public Time plus(int n, TUnit unit) {
		GregorianCalendar cal = getCalendar();
		cal.add(unit.getCalendarField(), n);
		return new Time(cal);
	}

	/**
	 * Convenience for {@link #plus(Dt)} where Dt = 1 unit
	 */
	public Time plus(TUnit dt) {
		return plus(dt.dt);
	}

	/**
	 * GMT time
	 */
	@SuppressWarnings("deprecation")
	@Override
	public String toString() {
//		GregorianCalendar cal = getCalendar();
//		if (cal.get(Calendar.ERA) == GregorianCalendar.BC) {
//			// TODO do something sensible here
//		}
		return getDate().toGMTString();
	}

	/**
	 * Inefficient convenience for using {@link SimpleDateFormat}.
	 * 
	 * @param dateFormat
	 *            E.g. "EEE, MMM d, yyyy" for "Wed, Jul 4, 2001", "h:mm a" for
	 *            12:08 PM. See {@link SimpleDateFormat} for details.
	 */
	public String toString(String dateFormat) {
		SimpleDateFormat sdf = new SimpleDateFormat(dateFormat);
		return sdf.format(getDate());
	}

}
